'''
simono

a fake object that manages the game state and pygame
'''

#stdlib
import time
import math
from random import randint
#lib
import pygame
from pygame.locals import *
#custom
import data
from util import * # ja..
import gameobjects
import explosions

SCREENWIDTH = 800
SCREENHEIGHT = 600


def main():
	print "Hello from my game's main()"
	noa = NotAnObject()
	noa.gameLoop()

class MixerControl(Borg):
	sound = "off" # on/off
	maxSounds = 50
	sounds = []

	def __init__(self):
		pygame.mixer.set_num_channels(self.maxSounds)

	def play(self,sound):
		if self.sound == "off":
			return
		if len(self.sounds) > self.maxSounds:
			self.sounds[0].stop()
			self.sounds = self.sounds[1:]
		s = data.loadSound(sound)
		s.play(0)
		self.sounds.append(s)



class NotAnObject:

	def __init__(self):
		pygame.init() # magic
		self.window = pygame.display.set_mode( (SCREENWIDTH,SCREENHEIGHT)) # get a surface to draw on
		pygame.display.set_caption( 'simono a game' ) # set a window caption
		self.backgroundSrf  = pygame.Surface( self.window.get_size() ) # create a surface, same size as window
		self.backgroundSrf.fill( (191,191,191) ) # fill surface with black color
		self.window.blit(self.backgroundSrf, (0,0) ) # blit (=put) background on window
		self.gameFont=pygame.font.SysFont(pygame.font.get_default_font(), 40)
		pygame.display.flip() # magic, needed to display window updates

		self.running = 1
		self.lastFrametime = 0
		self.fps = 70.0 # target fps

		self.helpImage = data.loadImage("help.png")

		self.state = "WAIT" # WAIT => SHRINK => DROP
		self.stateWaitStart = time.time()
		self.stateWaitDuration = 10

		# for dropping bottom part
		self.dropWidth = SCREENWIDTH
		self.dropHeight = None # different.. depend on where lowest plattform is

		# for falling top part
		self.fallHeight = 0
		self.startUp()

		self.mixer = MixerControl()

	def startUp(self):
		"""
		sets all important info, that must be present before startup
		can also be used for a complete reset
		"""
		self.actor = None
		self.obstacles = []
		self.bullets = []
		self.explosions = []
		self.enemies = []
		self.boxes = []
		self.state = "WAIT"
		self.stateWaitStart = time.time()


		# counting for score
		self.counter = {"enemykill":0,"boxkill":0,"cash":0,"drop":0}

		# enemies, obstacles
		lasti = 0
		i = 100
		while i <= SCREENHEIGHT-150:
			# important: -30 hier sind min abstand von zwei levels!!!!!
			(enemies,obstacles,boxes) = self.createNewLevel( (lasti,i-30))
			# on first starat we have bottom obstacle
			# create bottom obstacle
			self.obstacles.append( gameobjects.Obstacle(0,SCREENHEIGHT-11,SCREENWIDTH,11))
			self.obstacles += obstacles
			self.enemies += enemies
			self.boxes += boxes
			lasti = i
			i += randint(50,150)
		self.actor = gameobjects.Actor(SCREENWIDTH-50,SCREENHEIGHT-50)

		# IMPORATNT!!
		self.calcDropHeight() # set dropheight




	def gameLoop(self):
		"""
		this is UGLY.
		dont look at me. if you want to understand how i work, plesae take
		a look at gameobjects, makes more sense
		"""


		testStart = Vector(300,300)
		testVec = Vector(5,50)
		self.lastFrametime = time.time()
		cc = 0 # for frame printing
		fpslist = []
		while self.running == 1:
			fTime = time.time()
			nowtime = fTime
			timeElapsed = fTime - self.lastFrametime
			if timeElapsed > 1/self.fps:
				self.lastFrametime = fTime
				if cc > 60:
					fps = sum(fpslist) / 60.0
					fpslist = []
					print "FPS "+str(fps)
					cc = 0
				else:
					fpslist.append(1/timeElapsed)
					cc += 1
				# clear screen
				self.window.blit(self.backgroundSrf, (0,0) ) # blit (=put) background on window
				# OBSTACLES draw
				for o in self.obstacles:
					o.draw(self.window)

				# BOXES draw+update
				for b in self.boxes:
					# boxes <-> obstacles
					for o in self.obstacles:
						if o.rect.colliderect(b.getRect()):
							b.stuck = 1
					b.update(timeElapsed)
					b.draw(self.window)
					# no boxes<->bullets if its health
					if isinstance(b,gameobjects.HealthBox) or isinstance(b,gameobjects.CashBox):
						# healthboxes <-> actor
						if self.actor.getRect().colliderect(b.getRect()):
							self.boxes.remove(b)
							if isinstance(b,gameobjects.HealthBox):
								self.actor.health += 3
							else:
								self.counter["cash"] += 1
						continue

					# boxes <-> bullets
					for bull in [xb for xb in self.bullets if xb.stuck == 0]:
						if b.getRect().collidepoint(bull.position.x,bull.position.y):
							killBox = b.hitByBullet(bull)
							bull.stuck = 1
							self.bullets.remove(bull)
							if killBox:
								self.boxes.remove(b)
								# create an explosion
								self.explosions.append( explosions.BoxExplosion( b.position) )
								# add counter if bullet is from player
								if bull.type == 0:
									self.counter["boxkill"] += 1
								break
							else:
								self.explosions.append( explosions.WoodExplosion( b.position) )


				# EXPLOSIONS update, draw
				for e in self.explosions:
					e.update(timeElapsed)
					e.draw(self.window)
					if e.stuck == 1:
						self.explosions.remove(e)

				# USER INPUT
				# actor update
				self.handleKbdMouse()
				self.actor.update(timeElapsed)

				# enemies <-> bullets (both must be non-stuck)
				for b in [xb for xb in self.bullets if xb.stuck == 0 and xb.type == 0]:
					for e in [xe for xe in self.enemies if xe.stuck == 0]:
						if e.getRect().collidepoint(b.position.x,b.position.y):
							# kill enemy
							killEnemy = e.hitByBullet()
							if killEnemy:
								self.explosions.append( explosions.HumanExplosion( b.position))
								self.mixer.play("dying.wav")
								# add counter if bullet is from player
								if b.type == 0:
									self.counter["enemykill"] += 1
							else:
								self.explosions.append( explosions.BloodExplosion( b.position))
							b.stuck = 1
							self.bullets.remove(b)
							break

				# ENEMIES draw, update
				for e in self.enemies:
					newBullet = e.update(timeElapsed,self.actor,self.obstacles)
					if newBullet is not None:
						self.bullets.append(newBullet)
					e.draw(self.window)

				dropoff = SCREENHEIGHT - self.dropHeight # need this
					# dropoff... because obst dont collide if below this


				# actor <-> bullets
				for b in [xb for xb in self.bullets if xb.stuck == 0 and xb.type == 1]:
					if self.actor.getRect().collidepoint(b.position.x,b.position.y):
						self.actor.hitByBullet()
						self.explosions.append( explosions.BloodExplosion( b.position))
						b.stuck = 1
						self.bullets.remove(b)

				# rope <-> obstacle
				if self.actor.ropeStatus == "shoot":
					ropeHeadPosition = self.actor.getRopeHeadPosition()
					ropeHeadRect = pygame.Rect(ropeHeadPosition.x,
												ropeHeadPosition.y,
												5,5)
					for o in self.obstacles:
						# if obst is dropping.. forget it
						if o.position.y >= dropoff and (self.state in ["SHRINK"]):
							continue
						if o.rect.colliderect( ropeHeadRect) == True:
							self.actor.ropeCollideUpdate( ropeHeadPosition)
							break
				# actor <-> obstacle
				elif self.actor.ropeStatus == "none" and self.actor.obstacle is None:
					origRect = self.actor.getRect()
					hitHeight = 5
					hitWidth = 5
					actorRect = pygame.Rect(origRect.midbottom[0]-hitWidth/2.0,origRect.midbottom[1]-hitHeight/2.0,hitWidth,hitHeight)
					for o in self.obstacles:
						# if obst is dropping.. forget it
						if o.position.y >= dropoff and (self.state in ["SHRINK"]):
							continue
						topRect = pygame.Rect(o.rect.x,o.rect.y,o.rect.width,2)
						if topRect.colliderect(actorRect):
							self.actor.landedOnObstacle(o)
							break

				# bullets <-> obstacles
				for b in self.bullets:
					b.update(timeElapsed)
					for o in self.obstacles:
						if o.rect.collidepoint(b.position.x,b.position.y):
							b.stuck = 1
					b.draw(self.window)

				# fall, drop, managing
				if self.state == "WAIT":
					if nowtime - self.stateWaitStart >= self.stateWaitDuration:
						self.mixer.play("shrink.wav")
						self.state = "SHRINK"
						self.counter["drop"] += 1
					else:
						width = int(5 * ( (nowtime - self.stateWaitStart) / float(self.stateWaitDuration)))
						p1 = (0,SCREENHEIGHT-self.dropHeight)
						p2 = (SCREENWIDTH,SCREENHEIGHT-self.dropHeight)
						pygame.draw.line(self.window,(255,255,255),p1,p2,width)
				elif self.state == "SHRINK":
					self.shrinkBottom(timeElapsed)
				elif self.state == "DROP":
					self.dropDown(timeElapsed)

				# drop the actor of the obstacle, if he is standing on
				# on that is shrinking
				if self.state == "SHRINK":
					if (self.actor.obstacle is not None) and (self.actor.obstacle.position.y >= dropoff):
						self.actor.obstacle = None


				# draw actor - after bottom falling
				self.actor.draw(self.window)

				# help image
				self.window.blit(self.helpImage,(2,SCREENHEIGHT-145))

				# score
				score = (self.counter["enemykill"]*10 +
						self.counter["boxkill"]*2 +
						self.counter["cash"]*50+
						self.counter["drop"] * 5
						)
				score = "%08d"%score
				scoreSrf = self.gameFont.render(str(score),1,(0,0,0))
				self.window.blit(scoreSrf,(SCREENWIDTH-max(80,scoreSrf.get_rect().width),2))
				# healt bar
				self.actor.health  = min(100,self.actor.health ) # cap health
				health = self.actor.health * 1.5
				color = (0,255,0)
				if self.actor.health <= 35:
					color = (255,0,0)
				elif self.actor.health <= 60:
					color = (255,204,0)
				pygame.draw.rect(self.window,color,pygame.Rect(SCREENWIDTH-int(health)-2,scoreSrf.get_rect().height,int(health),20))
				pygame.draw.rect(self.window,(0,0,0),pygame.Rect(SCREENWIDTH-152,scoreSrf.get_rect().height,150,20),3)


				# check for death condition
				if self.actor.health <= 0:
					gameoverSrf = self.gameFont.render("Game Over",1,(0,0,0))
					grect = gameoverSrf.get_rect()
					self.window.blit(gameoverSrf,((SCREENWIDTH/2.0)-(grect.width/2.0),SCREENHEIGHT/2.0 - (grect.height/2.0)))
				pygame.display.flip()




#--- drop and shrinkage

	def shrinkBottom(self,timeElapsed):
		"""
		cut the lower part of the screen and let it shrink
		should look like falling into the screen
		"""
		if self.dropWidth <= 50:
			self.dropWidth = SCREENWIDTH
			self.bottomKillLevel()
			self.topCreateLevel()
			self.mixer.play("drop.wav")
			self.state = "DROP"
			return 1
		dropHeight= int(self.dropHeight) # part of orig screen we are shrking
		self.dropWidth -= int( (200*timeElapsed) )# width after shrink
		newDropWidth  = self.dropWidth
		newDropHeight = int(0.9* (newDropWidth/float(SCREENWIDTH)) * dropHeight )# height after shrink
		blackSurface = pygame.Surface((SCREENWIDTH,dropHeight)) # black for screen overdraw
		origDropTopLeft = (0,SCREENHEIGHT-dropHeight) # pos of drop area in orig screen
		temp= pygame.Surface((SCREENWIDTH,dropHeight)) # shrinkable surface
		temp.blit(self.window,(0,0),pygame.Rect((0,SCREENHEIGHT-dropHeight,SCREENWIDTH,dropHeight)))
		temp = pygame.transform.scale(temp,(newDropWidth,newDropHeight))
		self.window.blit(blackSurface,origDropTopLeft)
		newDropTopLeft = ( origDropTopLeft[0] + (SCREENWIDTH - newDropWidth)/2.0,
							origDropTopLeft[1] )
		self.window.blit(temp,newDropTopLeft)
		# draw wall lines
		lines = []
		lines.append( ( (0,SCREENHEIGHT),(newDropTopLeft[0],newDropTopLeft[1]+newDropHeight) ) )
		lines.append(
					  ( (SCREENWIDTH,SCREENHEIGHT),(newDropTopLeft[0]+newDropWidth,newDropTopLeft[1]+newDropHeight))
					)

		for a,b in lines:
			pygame.draw.aaline(self.window,(255,255,255),a,b,3)
		pygame.draw.rect(self.window,(255,255,255),pygame.Rect(newDropTopLeft,(newDropWidth,newDropHeight)),3)
		return 1

	def bottomKillLevel(self):
		"""
		kill off all objects that where in the falling part at the bottom
		"""
		dropoff = SCREENHEIGHT - self.dropHeight
##		print "killing after "+str(dropoff)
		objectlists = {"obstacle":self.obstacles,
						"box":self.boxes,
						"enemy":self.enemies,
						"bullet":self.bullets,
						"explosion":self.explosions}

		for oname,olist in objectlists.items():
			newlist = []
			for o in olist:
				if o.position.y <= dropoff:
					newlist.append(o)
			if oname == "obstacle":
				self.obstacles = newlist
			elif oname == "box":
				self.boxes = newlist
			elif oname == "enemy":
				self.enemies = newlist
			elif oname == "bullet":
				self.bullets = newlist
			elif oname == "explosion":
				self.explosions = newlist


		# kill actor
		if self.actor.position.y >= dropoff:
			self.actor.health = 0
		return 1


	def topCreateLevel(self):
		"""
		create new objects over screen
		"""
		if self.dropHeight < 50:
##			print "didnt create new level.. not enogh space"
			return 1
		(enemies,obstacles,boxes) = self.createNewLevel( (0-self.dropHeight,-50))
		self.obstacles += obstacles
		self.enemies += enemies
		self.boxes += boxes
##		print "created new level"
		return 1
	def dropDown(self,timeElapsed):

		"""
		move ALL OBJECTS down
		"""

		if self.fallHeight < self.dropHeight:
			fallamount = 30* timeElapsed
			if self.fallHeight + fallamount > self.dropHeight:
				fallamount = self.dropHeight - self.fallHeight
			self.fallHeight += fallamount
		else:
##			print "finished droppping"
			self.fallHeight = 0
			self.state = "WAIT"
			self.calcDropHeight()
			self.stateWaitStart = time.time()
			return 1

		objectlists = {"obstacle":self.obstacles,
						"box":self.boxes,
						"enemy":self.enemies,
						"bullet":self.bullets,
						"explosion":self.explosions}
		for oname,olist in objectlists.items():
			for o in olist:
				o.position[1] += fallamount
				if oname == "obstacle":
					o.recalcRect()
				if oname == "box":
					o.rect = None
					o.getRect()
		# extra code to drop the actor
		self.actor.position[1] += fallamount
		if self.actor.ropeHeadPosition is not None:
			self.actor.ropeHeadPosition[1] += fallamount

		# print black area
		blackheight = self.dropHeight - self.fallHeight
		blackSurface = pygame.Surface((SCREENWIDTH,blackheight)) # black for screen overdraw
		self.window.blit( blackSurface, (0,SCREENHEIGHT-blackheight))

		# todo: print white lines - ie copy surface from shrink


#--- helpers

	def createNewLevel(self,yinterval):
		(miny,maxy) = yinterval
##		print "new level for "+str(yinterval)
		y = randint(int(miny),int(maxy))
		levelMaxWidth = 800
		obstDone = 0
		obstMin = 80 # min width obstacle
		obstMax = 200 # max width
		obstHeight = 11 # obst height
		spaceMin = 30 # space between 2 obstacles min
		spaceMax = 200 # space between 2 obstacles max
		newObstacles = []
		cx = 0
		# create obstacles at one y pos
		while obstDone == 0:
			cx += randint(spaceMin,spaceMax)
			w = randint(min(obstMin,abs(levelMaxWidth-cx)),min(abs(levelMaxWidth-cx),obstMax))
			newObstacles.append( gameobjects.Obstacle( cx,y,w, obstHeight))
			cx += w
			if cx + obstMax + 5 >= levelMaxWidth:
				obstDone = 1

		# create enemies and boxes on the obstacles
		minEnemySpace = obstMin-10 # min pixels an enemy needs on an obstacle
		minBoxSpace = obstMin-30	# min pixels a box needs per obstacle
		minHealthBoxSpace = obstMin
		newEnemies = []
		newBoxes = []
		for ob in newObstacles:
			cx = ob.position.x
			# enemies
			for i in range(randint(0,round(ob.w/minEnemySpace))):
				cx += randint(0,min(minEnemySpace,(ob.position.x+ob.w-cx)))
				newEnemies.append( gameobjects.Enemy(Vector(cx,y-gameobjects.Enemy.height),ob))
			# boxes
			for i in range(randint(0,round(ob.w/minBoxSpace))):
				cx += randint(0,min(minBoxSpace,(ob.position.x+ob.w-cx)))
				newBoxes.append( gameobjects.Box(Vector(cx,y-gameobjects.Box.height-1),ob))
			# healthboxes
			if randint(0,5) == 5:
				cx += randint(0,min(minHealthBoxSpace,(ob.position.x+ob.w-cx)))
				newBoxes.append( gameobjects.HealthBox(Vector(cx,y-gameobjects.HealthBox.height-1),ob))
			# cash boxes
			if randint(0,8) == 8:
				cx += randint(0,min(minHealthBoxSpace,(ob.position.x+ob.w-cx)))
				newBoxes.append( gameobjects.CashBox(Vector(cx,y-gameobjects.CashBox.height-1),ob))



		return (newEnemies,newObstacles,newBoxes)

	def calcDropHeight(self):
		"""
		stor dropHeight for later use
		"""
		lowest = None
		for o in self.obstacles:
			if lowest is None or (o.position.y > lowest.position.y):
##				print "lowest is "+str(o.position.y)
				lowest = o
		self.dropHeight = int(SCREENHEIGHT - lowest.position.y + 50)


	def handleKbdMouse(self):
		"""
		handle keyboard and mouse events
		this includes creating nets in the self.nets list when necessary
		## see http://www.pygame.org/docs/ref/event.html
		"""
		for event in pygame.event.get():
			if event.type == QUIT: # user closed the window
				self.running = 0 # stop game loop
			elif event.type == KEYDOWN:
				if event.key == K_ESCAPE:
					self.running = 0 # stop game loop
				elif event.key == K_SPACE:
					self.actor.dropRopeRequest()
				elif event.key == K_RETURN:
					self.startUp()
				elif event.key in (K_w,K_a,K_s,K_d):
					self.actor.keyDown(event.key)
			elif event.type == KEYUP:
				if event.key == K_SPACE:
					self.actor.shortenRope = 0
				elif event.key in (K_w,K_a,K_s,K_d):
					self.actor.keyUp(event.key)
			elif event.type == MOUSEBUTTONDOWN:
				pass
			elif event.type == MOUSEBUTTONUP:
				# but3 == shoot the rope
				if event.button == 3:
					self.actor.shootRopeRequest(event.pos)
				elif event.button == 1:
					newbullet = self.actor.shootBulletRequest(event.pos)
					self.mixer.play("gun.wav")
					self.bullets.append(newbullet)


